#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <alloca.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "tools.h"

#define EXZESS_VENDOR ((uint16_t)0x0815)
#define EXZESS_DEVICE ((uint16_t)0x4711)

#define XILINX_VENDOR ((uint16_t)0x10ee)
#define XILINX_DEVICE ((uint16_t)0x0007)

int main(int argc, char **argv)
{
	set_log_level(LL_TRACE);
	char *dev_dir = find_pci_device_dir(EXZESS_VENDOR, EXZESS_DEVICE);
	if (! dev_dir) {
		dev_dir = find_pci_device_dir(XILINX_VENDOR, XILINX_DEVICE);
	}
	exit_if (NULL == dev_dir, "no suitable device found\n");
	log_msg(LL_NOTICE, "found device at %s\n", dev_dir);

	perror_if(-1 == chdir(dev_dir), "could not cd to %s", dev_dir);
	int resfd = open("resource0", O_RDWR);
	perror_if(-1 == resfd, "could not open resource0 file: ");

	uint32_t * volatile bar0 = mmap(NULL, 1024, PROT_READ | PROT_WRITE, MAP_SHARED, resfd, 0);
	perror_if(MAP_FAILED == bar0, "mmap of BAR 0 failed\n");

	// initiator reset
	bar0[0] = 1;
	bar0[0] = 0;

	uint32_t volatile dma_cr = bar0[1];
	// DMA control register
	printf("cr: %04x\n", dma_cr);

	// set TLP count to write from card
	bar0[4] = 5;

	// set write TLP payload size
	bar0[3] = 4;

	// set TLP count to read from card
	bar0[9] = 5;
	
	// set read TLP payload size
	bar0[8] = 4;

	// set read/write data patterns
	bar0[6] = 0xaa00aa00;
	bar0[5] = 0x08154711;

	// dma control register, read it, switch on the enable bits, write it to start transfers
	//
	dma_cr = bar0[1];
	/*                read enabl   write enab */
	dma_cr = dma_cr | 0x00010000 | 0x00000001;

	bar0[1] = dma_cr;
	dma_cr = bar0[1];

	printf("cr: %04x\n", dma_cr);

#if 0
	bar0[20] = 0xffffffff;
	bar0[21] = 0xffffffff;
#endif

	printf("%s", hexdump(bar0, 0x80));

	exit(EXIT_SUCCESS);
}
