#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <alloca.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <stdlib.h>

#include "tools.h"

#define EXZESS_VENDOR ((uint16_t)0x0815)
#define EXZESS_DEVICE ((uint16_t)0x4711)

#define XILINX_VENDOR ((uint16_t)0x10ee)
#define XILINX_DEVICE ((uint16_t)0x0007)

static char *find_pci_device_dir(uint16_t vendor, uint16_t device)
{
	char *sysfsdir = "/sys/bus/pci/devices";
	char *res = NULL;
	DIR *d;
	d = opendir(sysfsdir);
	char vendor_str[8];
	char device_str[8];
	sprintf(vendor_str, "0x%04x", vendor);
	sprintf(device_str, "0x%04x", device);

	exit_if(NULL == d, "could not open %s", sysfsdir);

	struct dirent *de;

	while (de = readdir(d)) {
		if (strcmp(".", de->d_name) == 0 || strcmp("..", de->d_name) == 0)
			continue;

		res = malloc(strlen(sysfsdir) + strlen(de->d_name) + 2);
		sprintf(res, "%s/%s", sysfsdir, de->d_name);

		char *vfn = alloca(strlen(sysfsdir) + strlen(de->d_name) + sizeof("vendor") + 3);
		char *dfn = alloca(strlen(sysfsdir) + strlen(de->d_name) + sizeof("device") + 3);
		sprintf(vfn, "%s/%s/%s", sysfsdir, de->d_name, "vendor");
		sprintf(dfn, "%s/%s/%s", sysfsdir, de->d_name, "device");

		errno = 0;
		int vfd = open(vfn, O_RDONLY);
		perror_if(0 == vfd, "failed to open vendor id file at %s", vfn);
		errno = 0;
		char vfs[10];
		perror_if(1 > read(vfd, vfs, 9), "read failed on vendor id file %s", vfn);
		vfs[6] = 0;

		errno = 0;
		int dfd = open(dfn, O_RDONLY);
		perror_if(0 == dfd, "failed to open device id file at %s", dfn);
		errno = 0;
		char dfs[10];
		perror_if(1 > read(dfd, dfs, 9), "read failed on vendor id file %s", dfn);
		dfs[6] = 0;

		perror_if(-1 == close(vfd) || -1 == close(dfd), "close failed");

		log_msg(LL_TRACE, "device at %s: %s:%s\n", res, vfs, dfs);

		if (strcmp(vfs, vendor_str) == 0 && strcmp(dfs, device_str) == 0) {
			/* we found it, clean up and return res */
			break;
		} else {
			/* nothing there, next one, or at the end, return res = NULL */
			free(res);
			res = NULL;
		}
	}
	closedir(d);

	return res;
}

int main(int argc, char **argv)
{
	set_log_level(LL_TRACE);
	char *dev_dir = find_pci_device_dir(EXZESS_VENDOR, EXZESS_DEVICE);
	if (! dev_dir) {
		dev_dir = find_pci_device_dir(XILINX_VENDOR, XILINX_DEVICE);
	}
	exit_if (NULL == dev_dir, "no suitable device found\n");
	log_msg(LL_NOTICE, "found device at %s\n", dev_dir);

	perror_if(-1 == chdir(dev_dir), "could not cd to %s", dev_dir);
	int resfd0 = open("resource0", O_RDWR);
	perror_if(-1 == resfd0, "could not open resource0 file: ");

	uint32_t * volatile bar0 = mmap(NULL, 1024, PROT_READ | PROT_WRITE, MAP_SHARED, resfd0, 0);
	perror_if(MAP_FAILED == bar0, "mmap of BAR 0 failed\n");

	int resfd1 = open("resource1", O_RDWR);
	perror_if(-1 == resfd1, "could not open resource1 file: ");

	uint32_t * volatile bar1 = mmap(NULL, 64 * 1024, PROT_READ | PROT_WRITE, MAP_SHARED, resfd1, 0);
	perror_if(MAP_FAILED == bar1, "mmap of BAR 1 failed\n");

	printf("%s", hexdump(bar0, 0x80));
	printf("%s", hexdump(bar1, 0x80));

	exit(EXIT_SUCCESS);
}
